using UnityEngine;
using System.Collections;

public class GameFlow : MonoBehaviour {
	
	public Material[] creatureMaterials; // All the possible creatues as separate materials
	public Transform[] spawnPoints;
	public GameObject creaturePrefab;
	public GameObject player;
	
	public GUIStyle countStyle;
	public GUIStyle timeStyle;
	public GUIStyle victimMessageStyle;
	public GUIStyle simpleMessageStyle;
	
	public Texture crosshairTexture;
	public Texture boxBackground;
	
	public AudioSource crowdAudioSource;
	public AudioSource clockAudioSource;
	public AudioSource messageAudioSource;
	public AudioSource narratorAudioSource;
	
	public AudioClip killSound;
	public AudioClip wrongChoice;
	public AudioClip clockTick;
	public AudioClip clockTock;
	public AudioClip laugh;
	public AudioClip sigh;
	
	private int hour = 0;
	private int time = 10;
	
	private ArrayList suspects;
	private Suspect murderer;
	private Suspect latestVictim;
	
	private string simpleMessage = "";
	private string victimMessage = "";
	private string endGameMessage = "";
	
	private bool gameLost = false;
	private bool gameWon = false;
	
	void OnEnable() {
		Messenger<Suspect>.AddListener("accuse", accuse);
	}
	
	void OnDestroy() {
	   	Messenger<Suspect>.RemoveListener("accuse", accuse);
	}
	
	void Start () {
		spawnCreatures(); 
		startCountDown();
	}
	
	void Update () {
		Screen.lockCursor = true;	
		
		// Check for game end state
//		if(gameWon || gameLost){
//			endGameMessage = gameWon ? "You win!" : "You lose!";
//			return;
//		}		
		
		if(time <= 0){
			time = 10;
			nextRound();
		}
		
		if(Input.GetButtonUp("Backspace")){
			Application.LoadLevel("main");
		}
		
		if(Input.GetButtonUp("E") || Input.GetButtonUp("Jump")){
			// If we're showing a simple message, dismiss it
			if(simpleMessage != ""){
				simpleMessage = "";
			}			
			// If we're showing a victim message, dismiss it and restart the counter
			else if(victimMessage != ""){
				victimMessage = "";
				latestVictim = null;
				if(!gameWon && !gameLost)
					startCountDown();
				freezePlayer(false);
			} 
			// Restart game... or something
			else if(endGameMessage != ""){
				Application.LoadLevel("main");
			}
		}
		
		if(suspects.Count <= 1 && !gameLost && !gameWon){
			loseGame();
		}
	}
	
	void OnGUI() {
		// Adjust the font size
		simpleMessageStyle.fontSize = Screen.width/22;
		victimMessageStyle.fontSize = Screen.width/22;
		countStyle.fontSize = Screen.width/25;
		timeStyle.fontSize = Screen.width/25;
		
		GUI.DrawTexture(new Rect(Screen.width / 2 - (crosshairTexture.width / 2), Screen.height / 2 - (crosshairTexture.height / 2), crosshairTexture.width, crosshairTexture.height), crosshairTexture);
		GUI.Label(new Rect(10, 10, Screen.width - 20, 150), "Cycle " + hour, countStyle);
		GUI.Label(new Rect(10, 10, Screen.width - 20, 150), "Time " + time, timeStyle);
		
		if(simpleMessage != "" || victimMessage != "" || endGameMessage != ""){
			freezePlayer(true);
			
			// Message background
			GUI.DrawTexture(new Rect(0, 0, Screen.width, Screen.height), boxBackground, ScaleMode.ScaleToFit);
			
			// Show simple message
			if(simpleMessage != ""){
				GUI.Label(new Rect(Screen.width * 0.2f, Screen.height / 4 , Screen.width * 0.7f, Screen.height / 2), simpleMessage, simpleMessageStyle);
			}
			
			// Show the latest victim and message
			else if(latestVictim != null){
				Graphics.DrawTexture(new Rect(Screen.width * 0.65f, Screen.height * 0.2f, Screen.width / 4, Screen.height * 0.6f), latestVictim.getMaterial().mainTexture, latestVictim.getMaterial());
				GUI.Label(new Rect(Screen.width * 0.2f, Screen.height / 4 , Screen.width * 0.45f, Screen.height / 2), victimMessage, simpleMessageStyle);
			}
			// Show end of game message if we have one
			else if(endGameMessage != ""){
				GUI.Label(new Rect(Screen.width * 0.2f, Screen.height / 4 , Screen.width * 0.7f, Screen.height / 2), endGameMessage, simpleMessageStyle);
			}
		}
	}

	
	//------------------------------------------------------------
	
	private void spawnCreatures(){
		ArrayList materialList = new ArrayList();
		materialList.AddRange(creatureMaterials);
		
		suspects = new ArrayList();
		
		// set a unique creature for each spawn point
		for(int n = 0 ; n < spawnPoints.Length ; n++){
			// get random material
			int rand = Random.Range(0, materialList.Count);
			Material mat = materialList[rand] as Material;
			materialList.Remove(mat);
			
			// Set material on prefab and instantiate
			creaturePrefab.transform.Find("plank").renderer.material = mat;
			GameObject creature = Instantiate(creaturePrefab, spawnPoints[n].position, spawnPoints[n].rotation) as GameObject;
			suspects.Add(creature.GetComponent("Suspect") as Suspect);
		}
		
		// Choose a murderer
		murderer = suspects[Random.Range(0, suspects.Count)] as Suspect;
	}
	
	/**
	 * TODO This would have been a lot simpler with a sorted list...
	 */
	private void killSomeone(){
		// Working list, all suspects minus the killer
		ArrayList tempSuspectList = new ArrayList();
		tempSuspectList.AddRange(suspects);
		tempSuspectList.Remove(murderer);
		
		/*
		
		// Go through all suspects at random and try to kill one
		while(tempSuspectList.Count > 0){
			Suspect candidate = tempSuspectList[Random.Range(0, tempSuspectList.Count)] as Suspect;
			if(candidate.strength <= murderer.strength){
				// KILL
				Debug.Log(murderer.strength + ">=" + candidate.strength + " - Killed " + candidate.species + "-" + candidate.strength);
				murder(candidate);
				tempSuspectList.Clear();				
				return;
			} else {
				// SKIP
				//Debug.Log(murderer.strength + "<" + candidate.strength + " - No good, next!");
			}
			tempSuspectList.Remove(candidate);
		}
		
		*/
		
		// If none could be killed, kill a random strongest or equal
		// (we only reach this point if the above loop didn't return)
		if(suspects.Count <= 1){
			Debug.Log("Only the murderer left!");
			return;
		}

		// Go through all suspects at random and kill the weakest one possible
		Suspect weakestCandidate = tempSuspectList[Random.Range(0, tempSuspectList.Count)] as Suspect;;
		tempSuspectList.Remove(weakestCandidate);
		while(tempSuspectList.Count > 0){
			Suspect candidate = tempSuspectList[Random.Range(0, tempSuspectList.Count)] as Suspect;
			if(candidate.strength < weakestCandidate.strength){
				weakestCandidate = candidate;
			}
			tempSuspectList.Remove(candidate);
		}
		// Kill the weakest if we found one (and only if weaker then the killer), and return, otherwise we fall through to next loop
		if(weakestCandidate.strength < murderer.strength){
			murder(weakestCandidate);
			return;
		}
		
		tempSuspectList.AddRange(suspects);
		tempSuspectList.Remove(murderer);
		Suspect strongestCandidate = tempSuspectList[Random.Range(0, tempSuspectList.Count)] as Suspect;;
		tempSuspectList.Remove(strongestCandidate);
		while(tempSuspectList.Count > 0){
			Suspect candidate = tempSuspectList[Random.Range(0, tempSuspectList.Count)] as Suspect;
			if(candidate.strength >= strongestCandidate.strength){
				strongestCandidate = candidate;
			}
			tempSuspectList.Remove(candidate);
		}
		//Debug.Log("Strongest candidate: " + strongestCandidate.strength + " - Killed " + strongestCandidate.species + "-" + strongestCandidate.strength);
		murder(strongestCandidate);
	}
	
	/**
	 * Kill the given suspect and mark as latest victim
	 */
	private void murder(Suspect victim){
		messageAudioSource.clip = killSound;
		messageAudioSource.Play();
		latestVictim = victim;
		victimMessage = "My congregation shrinks with every hesitation.\n\nYour indecisiveness has cost us another life:";
		kill (victim);
		stopCountDown();
	}
	
	/**
	 * Just remove the victim from the game
	 */
	private void kill(Suspect victim){
		suspects.Remove(victim);
		
		// Only one left and we have't found the murderer? game over
		if(suspects.Count <= 1){
			loseGame();
		}
		
		victim.makeInvisible();
		//victim.enabled = false;
	}
	
	private void nextRound(){
		hour++;
		killSomeone();
	}
	
	private void tick(){
		if(time % 2 == 0){
			clockAudioSource.clip = clockTick;
		} else {
			clockAudioSource.clip = clockTock;
		}
		clockAudioSource.Play();
		time--;
	}
	
	private void startCountDown(){
		time = 10;
		InvokeRepeating("tick", 1, 1);
	}
	
	private void stopCountDown(){
		CancelInvoke("tick");
	}
	
	private void freezePlayer(bool freeze){
		MouseLook mlook = player.GetComponent("MouseLook") as MouseLook;
		mlook.enabled = !freeze;
		MouseLook camLook = Camera.main.GetComponent("MouseLook") as MouseLook;
		camLook.enabled = !freeze;
		player.GetComponent<FPSInputController>().enabled = !freeze;
		player.GetComponent<CharacterMotor>().enabled = !freeze;
	}
	
	private void accuse(Suspect accused){
		// Don't accuse anyone if we're showing a message
		if(victimMessage != "" || simpleMessage != ""){
			return;
		}
		
		stopCountDown();
		if(accused.Equals(murderer)){
			string cycStr = hour == 1 ? " cycle." : " cycles.";
			simpleMessage = "You have chosen wisely, little one.\n\nThe culprit was found after " + hour + cycStr + "\n\nPray that your hesitation was worth it. " + (suspects.Count-1) + " remain.";
			winGame();
		} else {
			messageAudioSource.Stop();
			messageAudioSource.clip = wrongChoice;
			messageAudioSource.Play();
			simpleMessage = "You chose... unwisely.\n\nThat creature did not admit to being the culprit.";
			kill(accused);
			//killSomeone();
			nextRound();
		}
		
	}
	
	private void loseGame(){
		Debug.Log("Game lost");
		stopCountDown();
		narratorAudioSource.Stop();
		narratorAudioSource.clip = sigh;
		narratorAudioSource.Play();
		gameLost = true;
		endGameMessage = "You failed to find the culprit in time. But despair not, little one.\nAn unfit inquisitor may yet make a worthy sacrifice.\n(press space to restart)";
	}
	
	private void winGame(){
		Debug.Log("Game won");
		stopCountDown();
		narratorAudioSource.Stop();
		narratorAudioSource.clip = laugh;
		narratorAudioSource.Play();
		gameWon = true;
		endGameMessage = "Press space to restart";
	}
	
}

